/*
File: CTO.CPP
Description: Code Table Optimization Compression-Decompression Implementation
Copyright 2000, Ian Fernandes
File Format:	Byte0	1	2	3	4				5........           ...
					I	C	A	C   NoOfSymbols-1	List Of Symbols 	Data
*/

#include <io.h>
#include <string.h>
#include <iostream.h>
#include <errno.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <Time.h>
#include <Conio.h>
#include <math.h>
#include "BitIO.CPP"

#define BYTES_TO_READ	1024
#define _FINAL_
typedef unsigned int UINT;
typedef unsigned long ULONG;
typedef unsigned char BYTE;

UINT NoOfSymbols;

char *Input;
char *Output;

FILE *hfInput, *hfOutput;
int _DEBUG_ =1;


struct SYM_TABLE
{
	unsigned char Symbol;
	unsigned int CodeLen;
	BYTE Code;
	SYM_TABLE *Next;
} *SymTable;

int Count[256];

void PrintUsage()
{
	cout <<"\nCode Table Optimization Compression-Decompression Agent";
	cout <<"\nCopyright 2000, Ian Fernandes";
	cout <<"\nUsage:cto inputfile outputfile";
}
//Chars to read determined by BYTES_TO_READ

void CountSymbols()
{
	unsigned char cbuff;
	for(int cntr=0;cntr<=255;cntr++)
		Count[cntr]=0;
	BitIO inp;
	fseek(hfInput, 0 , SEEK_SET);
	if(inp.Start(hfInput, (BOOL)0, (BOOL)1))
	{
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}
	cout <<"\nCounting Symbols[                    ]\r";
	cout <<"Counting Symbols[";

	long charsread, mult;
	int handle;
	long filelen= filelength(handle = (open(Input,O_RDONLY)));
	close (handle);
	mult =0;
	charsread=0;
	while(inp.StreamIn(&cbuff,8) != -1)
	{
		++Count[cbuff];
		++charsread;
		if(charsread > filelen*mult/100)
		{
			mult+=5;
			cout <<".";
		}
	}
	inp.End();

}



int CTOCompress()
{
	ULONG TotalBytes=0;
	short TotalSymbols=0;
	short temp, Bits;
	short bitlen;
	char outbit, outbittemp;
	BYTE code;
	SYM_TABLE *SymStart;
	for (int cntr=0; cntr <=255;cntr++)
	{
//		cout <<" "<<(char)cntr<<"	"<<Count[cntr];
		if(Count[cntr] !=0)
			++TotalSymbols;
		TotalBytes +=Count[cntr];
	}
	cout <<"\nTotal:"<<TotalBytes<<" Symbols:"<<TotalSymbols;
	temp=0;
	Bits=0;
	while((pow(2,temp) < (int)TotalSymbols))
	{
		++Bits;
		++temp;
	 }
	cout <<"\nCode Bits:"<<Bits<<" Total Encoded symbol:"<<pow(2,temp);
	cout <<"\nNow Encoding:\n";
	code =0;
	for(cntr=0;cntr<=255;cntr++)
	{
		if(Count[cntr] !=0)
		{
			if(SymTable == NULL)
			{	SymTable= new SYM_TABLE;
			SymStart=SymTable;
			}
			else
			{
				SymTable->Next = new SYM_TABLE;
				SymTable = SymTable->Next;
			}

			SymTable->Symbol = (unsigned char)cntr;
			SymTable->CodeLen = Bits;
			SymTable->Code = (BYTE) code;
			SymTable->Next = 0;
			++code;
//			printf("%c %i\t",SymTable->Symbol,SymTable->Code);

		}
	}
	SymTable=SymStart;
	cout <<"\nPreparing to Write:";
	cout <<"\nWriting Header";

	fseek(hfOutput,0,SEEK_SET);
	if(fprintf(hfOutput,"ICAC%c",TotalSymbols-1) == EOF)
	{
		cout <<"\nError Writing";
		return 1;
	}
	cout <<"\nWriting Code Table";

	while(SymTable!=NULL)
	{
		fwrite(&SymTable->Symbol,sizeof(SymTable->Symbol),1,hfOutput);
		SymTable = SymTable->Next;
	}
	SymTable=SymStart;


//~~~~~~~~~~~~~~~~~~TEST FOR BitIO
#ifdef _TEST_
{	BitIO out,in;
	char codete,hdr[5], sym[255];
	out.Start(hfOutput, (BOOL)1,(BOOL) 1);
	while(SymTable!=NULL)
	{
		out.StreamOut(SymTable->Code, Bits);
		SymTable = SymTable->Next;
	}
	out.End();
	fseek(hfOutput,0,0);
	fread(hdr,1,5,hfOutput);
	fread(sym,1,hdr[4]+1,hfOutput);
	cout <<"\n"<<"FP"<<ftell(hfOutput)<<"\n";

	in.Start(hfOutput, (BOOL)0,(BOOL)1);
	while(in.StreamIn(&codete, Bits) != -1)
	{
		printf("%c %i\t", sym[codete],codete);
	}
	in.End();
}
#endif
//~~~~~~~~~~~~~~~~~~~~TEST END*/


/*	cout <<"\nWriting to file...";
	int charsread;
	char  cbuff[BYTES_TO_READ];
	BitIO out;
	if(out.Start(hfOutput, BOOL(1),BOOL(1)))
	{
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}
	fseek(hfInput,0,SEEK_SET);
	do
	{
		charsread = fread(cbuff, sizeof(char), BYTES_TO_READ,hfInput);
		for( int cntr =0; cntr<charsread;cntr++)
		{
			while(SymTable->Symbol != cbuff[cntr])
			{
				SymTable = SymTable->Next;
				if(SymTable == NULL)
				{
						cout <<"\nLogical error";
						fclose(hfInput);
						fclose(hfOutput);
						exit(1);
				}
			}
			if(out.StreamOut(SymTable->Code, SymTable->CodeLen))
			{
				fclose(hfInput);
				fclose(hfOutput);
				exit(1);
			}
			SymTable=SymStart;
		}
	}
	while (charsread !=0);
	if(out.End())
	{
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}
	cout <<"Done";
   */

//~~~~~~~~~~~~~~~~~~FOR BINARY
#ifdef _FINAL_
	unsigned char  cbuff;
	BitIO out, inp;
	if(out.Start(hfOutput, BOOL(1),BOOL(1)))
	{
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}
	fseek(hfInput,0,SEEK_SET);
	if(inp.Start(hfInput, BOOL(0),BOOL(1)))
	{
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}
	cout <<"\nWriting to file[                    ]\r";
	cout <<"Writing to file[";

	long charsread, mult;
	int handle;
	long filelen= filelength(handle = (open(Input,O_RDONLY)));
	close (handle);
	mult =0;
	charsread=0;

	while(inp.StreamIn(&cbuff, 8) != -1)
	{
		while(SymTable->Symbol != cbuff)
		{
			SymTable = SymTable->Next;
			if(SymTable == NULL)
			{
					cout <<"\nLogical error";
					fclose(hfInput);
					fclose(hfOutput);
					exit(1);
			}
		}
		if(out.StreamOut(SymTable->Code, SymTable->CodeLen))
		{
			fclose(hfInput);
			fclose(hfOutput);
			exit(1);
		}
		SymTable=SymStart;
		++charsread;
		if(charsread > filelen*mult/100)
		{
			mult+=5;
			cout <<".";
		}

	}

	if(out.End())
	{
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}
	if(inp.End())
	{
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}
#endif

	return 0;
}

int main(int argc, char *argv[])
{
	clock_t tstart, tend;
	tstart=clock();
	PrintUsage();
	if( argc < 3)
	{
		cout <<"\nInsufficient Args";
		exit(1);
	}

	Input = new char[sizeof(argv[1])];
	Input = argv[1];
	if( argc >2)
	{
		Output =new char[sizeof(argv[2])];
		Output = argv[2];
	}

//Open File On Success returns 1 on failure exits with code 1
	cout <<"\n";
	if ((hfInput=fopen(Input,"rb")) == NULL)
	{
		perror(Input);
		exit(1);
	}
	if(_DEBUG_)
		cout <<"\n"<<Input<<" opened successfully";


	if((hfOutput=fopen(Output,"w+b")) ==NULL)
	{
		fclose(hfInput);
		perror (Output);
		exit(1);
	}

	if(_DEBUG_)
		cout <<"\n"<<Output<<" opened successfully";


	CountSymbols();

	CTOCompress();

	fclose(hfInput);
	fclose(hfOutput);

	tend=clock();
	cout <<"\nTime:"<<(tend-tstart)/CLK_TCK<<"\n";
	return 0;
}