/*
File: CTOU.CPP
Description: Code Table Optimization Decompression Implementation
Copyright 2000, Ian Fernandes
File Format:	Byte0	1	2	3	4				5........           ...
					I	C	A	C   NoOfSymbols-1	List Of Symbols 	Data
*/

#include <io.h>
#include <string.h>
#include <iostream.h>
#include <errno.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <Time.h>
#include <Conio.h>
#include <math.h>
#include <string.h>
#include "BitIO.CPP"

#define BYTES_TO_READ	1024

typedef unsigned int UINT;
typedef unsigned long ULONG;
typedef unsigned char BYTE;

UINT NoOfSymbols;

char *Input;
char *Output;

FILE *hfInput, *hfOutput;
int _DEBUG_ =1;


struct SYM_TABLE
{
	unsigned char Symbol;
	unsigned int CodeLen;
	BYTE Code;
	SYM_TABLE *Next;
} *SymTable, *SymStart;

ULONG Count[256];
unsigned char Hdr[4];
unsigned char SymNo;

void PrintUsage()
{
	cout <<"\nCode Table Optimization Compression-Decompression Agent";
	cout <<"\nCopyright 2000, Ian Fernandes";
	cout <<"\nUsage:cto inputfile outputfile";
}

void ReadHeader()
{
	fseek(hfInput, 0, SEEK_SET);
	if(fread(Hdr, 1, 4, hfInput) !=4)
	{
		cout<<"\nUnable to read header";
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}
	if( strcmp(Hdr,"ICAC") !=0)
	{
		cout <<"\nNot ICAC archive";
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}
	cout <<"\nHeader is "<<Hdr<<" ";
}

void CreateSymTable()
{
	cout <<"FP POS "<<ftell(hfInput);
	if(	fread((unsigned char *)&SymNo, 1, 1, hfInput) != 1)
	{
		cout<<"Read Error";
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}

	cout <<"\nNo Of Symbols "<<(int)(SymNo+1);
	cout <<"\n";
	int cntr=0;
	int bitsize =0;

	while((pow(2,cntr) < (int)SymNo+1))
	{
		++bitsize;
		++cntr;
	 }
	for(cntr=0;cntr <=(int)SymNo;cntr++)
	{
		if(SymTable==NULL)
		{
			SymTable = new SYM_TABLE;
			SymStart = SymTable;
		}
		else
		{
			SymTable->Next = new SYM_TABLE;
			SymTable = SymTable->Next;
		}

		SymTable->Code =(unsigned char)cntr;
		SymTable->CodeLen = bitsize;
		if(fread(&SymTable->Symbol,1,1,hfInput) != 1)
		{
			cout <<"\nError Reading";
			fclose(hfInput);
			fclose(hfOutput);
			exit(1);
		}
		SymTable->Next =0;
//		printf("%c %i\t",SymTable->Symbol,SymTable->Code);

	}


}


void Decode()
{
	cout <<"\nFP POS "<<ftell(hfInput);
	BitIO inp, out;
	if(fseek(hfOutput,0,SEEK_SET))
	{
			fclose(hfInput);
			fclose(hfOutput);
			cout <<"\n";
			perror(Output);
			exit(1);
	}
	if(	inp.Start(hfInput,BOOL(0),BOOL(1)))
	{
			fclose(hfInput);
			fclose(hfOutput);
			exit(1);
	}
	if(	out.Start(hfOutput,BOOL(1),BOOL(1)))
	{
			fclose(hfInput);
			fclose(hfOutput);
			exit(1);
	}
	SymTable= SymStart;
	int bitsize = SymTable->CodeLen;
	unsigned char code, codeint;

	cout <<"\nDecompressing[                    ]\r";
	cout <<"Decompressing[";

	long bitsread, mult;
	int handle;
	long filelen= filelength(handle = (open(Input,O_RDONLY)));
	close (handle);
	mult =0;
	bitsread=0;

	while( inp.StreamIn(&code, bitsize) != -1)
	{
		while(SymTable->Code != code)
		{
			SymTable= SymTable->Next;
			if(SymTable ==NULL)
			{
				cout <<"\nLogical error";
				fclose(hfInput);
				fclose(hfOutput);

				exit(1);
			}
		}
//		fwrite(&SymTable->Symbol,1,1,hfOutput);
		if(out.StreamOut((int)SymTable->Symbol,8))
		{
			fclose(hfInput);
			fclose(hfOutput);
			exit(1);
		}
		SymTable=SymStart;

		bitsread += bitsize;
		if((bitsread/8.0) > filelen*mult/100)
		{
			mult+=5;
			cout <<".";
		}


	}
	cout <<"\nFP POS "<<ftell(hfInput);
	if(	inp.End())
	{
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}
	if(out.End())
	{
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}
}


int main(int argc, char *argv[])
{
	clock_t tstart, tend;
	tstart=clock();
	PrintUsage();
	if( argc < 3)
	{
		cout <<"\nInsufficient args";
		exit(1);
	}

	Input = new char[sizeof(argv[1])];
	Input = argv[1];
	if( argc >2)
	{
		Output =new char[sizeof(argv[2])];
		Output = argv[2];
	}

//Open File On Success returns 1 on failure exits with code 1
	if ((hfInput=fopen(Input,"rb")) == NULL)
	{
		cout <<"\n";
		perror(Input);
		exit(1);
	}
	if(_DEBUG_)
		cout <<"\n"<<Input<<" opened successfully";


	if((hfOutput=fopen(Output,"w+b")) ==NULL)
	{
		fclose(hfInput);
		cout <<"\n";
		perror (Output);
		exit(1);
	}

	if(_DEBUG_)
		cout <<"\n"<<Output<<" opened successfully";


	ReadHeader();
	CreateSymTable();
	Decode();
	fclose(hfInput);
	fclose(hfOutput);

	tend=clock();
	cout <<"\nTime:"<<(tend-tstart)/CLK_TCK<<"\n";
	return 0;

}
